/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * An implementation of vrpsx_allocator_t that tracks how it is used
 * for the purpose of testing code that should use that interface in
 * particular ways.
 */

#define __STRICT_ANSI__ 1

#include "test_allocator.h"
#include <vrpsx_utility.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define TEST_ALLOCATION_MAGIC_WORD 0x3fcfc377
#define TEST_ALLOCATION_MAGIC_CHAR ((char)0xDE)

#define ZERO(PTR)     (memset((PTR), 0, sizeof *(PTR)))


static int nul_terminated_allocation(
    test_allocation_t *allocation)
{
    if (allocation->size) {
        char final = ((char*)(allocation + 1))[allocation->size - 1];
        if (final != '\0' && final != TEST_ALLOCATION_MAGIC_CHAR) {
            return 0;
        }
    }
    return 1;
}


static vrpsx_allocator_end_t const allocator_end;

static void test_allocator_free_memory(
    vrpsx_allocator_t *allocator,
    void *memory)
{
    if (! memory) {
        return;
    }

    VRPSX_UPCAST_NONNULL_DECL(
        test_allocator_t, , obj, interface, allocator);

    test_allocation_t *allocation = (test_allocation_t *)
            ((char*)memory - sizeof(test_allocation_t));

    if (allocation->magic != TEST_ALLOCATION_MAGIC_WORD) {
        printf("%s: bad allocation magic\n", __func__);
        exit(1);
    }

    if (memcmp((char*)(allocation + 1) + allocation->size, &allocation->magic,
                    sizeof allocation->magic)) {
        printf("%s: allocation overrun\n", __func__);
        exit(1);
    }

    if (! nul_terminated_allocation(allocation)) {
        obj->nul_terminated = 0;
    }

    if (allocation->allocator != allocator) {
        printf("%s: wrong allocator\n", __func__);
        exit(1);
    }

    if (obj->current_allocations-- == 0) {
        printf("%s: too many deallocations\n", __func__);
        exit(1);
    }

    dlink_destroy(&allocation->element);

    memset(allocation, 0, sizeof allocation + allocation->size);
    free(allocation);
}

static void *test_allocator_allocate_memory(
    vrpsx_allocator_t *allocator,
    size_t size)
{
    VRPSX_UPCAST_NONNULL_DECL(
        test_allocator_t, , obj, interface, allocator);

    test_allocation_t *allocation = (test_allocation_t *)
        malloc(sizeof(test_allocation_t) + size + sizeof allocation->magic);

    ZERO(allocation);
    dlink_construct(&allocation->element);
    allocation->size = size;
    allocation->magic = TEST_ALLOCATION_MAGIC_WORD;
    allocation->allocator = allocator;
    memset((char*)(allocation + 1), TEST_ALLOCATION_MAGIC_CHAR, size);
    memcpy((char*)(allocation + 1) + size, &allocation->magic,
            sizeof allocation->magic);

    dlink_splice(&obj->allocations, &allocation->element);

    ++ obj->current_allocations;
    if (obj->maximum_allocations < obj->current_allocations) {
        obj->maximum_allocations = obj->current_allocations;
    }

    return allocation + 1;
}

static vrpsx_allocator_vtbl_t const test_allocator_vtbl = {
    &test_allocator_free_memory,
    &test_allocator_allocate_memory,
    &allocator_end
};

extern void test_allocator_construct(
    test_allocator_t *obj)
{
    ZERO(obj);
    obj->interface.vtbl = &test_allocator_vtbl;
    dlink_construct(&obj->allocations);
    obj->current_allocations = 0;
    obj->maximum_allocations = 0;
    obj->nul_terminated = 1;
}

extern void test_allocator_destroy(
    test_allocator_t *obj)
{
    while (obj->allocations.next != &obj->allocations) {
        test_allocation_t *allocation = VRPSX_UPCAST_NONNULL(
            test_allocation_t, element, obj->allocations.next);
        dlink_destroy(&allocation->element);
        free(allocation);
    }
    obj->interface.vtbl = NULL;
}

extern int assert_singular_allocation(
    where_t where,
    test_allocator_t *obj,
    void *block,
    size_t size)
{
    test_allocation_t *allocation;

    if (block) {
        if (obj->current_allocations != 1) {
            print_where(where);
            printf("expected just one current allocation: %lu\n",
                    (unsigned long)obj->current_allocations);
            return -1;
        }

        allocation = VRPSX_UPCAST_NONNULL(
            test_allocation_t, element, obj->allocations.next);

        if (block != (char *)allocation + sizeof(test_allocation_t)) {
            print_where(where);
            printf("wrong allocation address\n");
            return -1;
        }

        if (size != allocation->size) {
            print_where(where);
            printf("wrong allocation size: %lu instead of %lu\n",
                    (unsigned long)allocation->size, (unsigned long)size);
            return -1;
        }
    } else {
        if (obj->current_allocations != 0) {
            print_where(where);
            printf("expected no current allocation: %lu\n",
                    (unsigned long)obj->current_allocations);
            return -1;
        }
    }

    if (obj->maximum_allocations > 1) {
        print_where(where);
        printf("multiple simultaneous prior allocations: %lu\n",
                (unsigned long)obj->maximum_allocations);
        return -1;
    }

    return 0;
}


extern int assert_nul_terminated_allocation(
    where_t where,
    test_allocator_t *obj)
{
    int nul_terminated = obj->nul_terminated;
    dlink_t *node = &obj->allocations;
    test_allocation_t *allocation;

    while (nul_terminated) {
        node  = node->next;
        if (node == &obj->allocations) {
            break;
        }

        allocation = (test_allocation_t *)(
            (char *)node - offsetof(test_allocation_t, element));

        if (! nul_terminated_allocation(allocation)) {
            nul_terminated = 0;
        }
    }

    if (! nul_terminated) {
        print_where(where);
        printf("non-NUL write to final octet of allocation\n");
        return -1;
    }

    return 0;
}
