/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Defines an interface that can be presented by a dynamically
 * loadable library in order to declare its top-level functions.
 */

#ifndef VRPSX_IDENTIFIED_FUNCTION_H
#define VRPSX_IDENTIFIED_FUNCTION_H 1

#include "vrpsx_identified.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* A dynamically loadable library may lists various functions
 * provided by that library along with interface identifiers
 * that specify how those functions are called.
 *
 * For a library to use this particular interface to present such a list,
 * it must prefix an underscore to the standard text form of the identifier
 * for this interface (see "vrpsx_identified.h" for details) and define the
 * resulting C identifier as an array of this type:
 *
 *     vrpsx_identified_function_t const []
 *
 * This list must be in order of decreasing preference and must
 * be terminated by an element whose "function" member is NULL.
 *
 * The library should make the resulting array symbol externally visible
 * in whatever manner is appropriate to the operating environment.
 * In C++ you must arrange the identifier is 'extern "C"' in case
 * name mangling would change its link-level symbol.
 *
 * The idea is that a program will load the library, look up the array by name,
 * use it to pick the most-preferred function among those that it knows how to
 * call, and then call that function in the appropriate manner.
 *
 * Extension:
 *
 * If you want to write a program that can load libraries that make use
 * of this interface and also libraries that make use of a new search
 * interface that you define, then pick different linkage symbol(s) for the
 * new interface and search for those first, and if you do not find them,
 * then search for the symbol defined by this interface as a fall-back.
 *
 * But please note that you are actually defining two new interfaces: a narrow
 * interface that defines only the new linkage symbol(s), and a compound and
 * perhaps implicitly defined interface that supports both old and new symbols.
 * Only the compound interface would actually extend this interface.  This
 * distinction matters when you wish to phase out support for an old narrow
 * interface: there is no way to do so while extending an old compound
 * interface, but you can still extend other narrow interfaces directly.
 *
 * Alternatives:
 *
 * An alternative approach would be to extend vrpsx_identified_t with basic
 * error reporting functionality, and then ask loadable libraries to provide
 * a function that makes use of that interface to try to obtain others.  The
 * two approaches are to some extent redundant, but could also be combined.
 *
 * We make the mostly arbitrary choice to provide vrpsx_identified_function_t
 * based on a guess that in typical cases there will be less code this way.
 *
 * Generalization:
 *
 * We describe this interface in terms of dynamic loading of libraries, but
 * you might find other dynamic call situations in which it might find use.
 *
 * Client/Implementation Terminology and Extensions:
 *
 * Our convention is to refer to the library as the client of this interface
 * and each listed function interface, and to refer to the the program that
 * loads the library as the implementation of (some of) those interfaces.
 *
 * Though "client" and "implementation" often oversimplify the relationships
 * implied by any interface, this convention might seem more backward than is
 * usual.  Our rationale for this choice of terminology is that the expected
 * pattern of usage is for the called function to receive another interface as
 * one of its arguments, with the expectation that it will make calls to that
 * interface.  Our terminology avoids the need to mentally swap the terms
 * "client" and "implementation" once that call has been made.
 *
 * Similarly, our choice avoids contravariance in the "extension" relationship:
 * for example, a main-like function interface involving a system interface
 * function argument is extended by a program that can provide an extension
 * of that system interface, rather than being narrowed by dynamic library
 * that requires such a system interface extension.  We hope that terms like
 * "backward compatible" are less confusing with the roles defined in this way.
 */
typedef struct vrpsx_identified_function_struct {
    /* A function whose whose calling requirements constitute
     * an identified interface.  Typically you must cast the
     * type of this pointer before calling through it.
     *
     * By convention, a NULL function pointer indicates
     * the end of a list of function-identifier pairs.
     */
    void (*function)(void);

    /* The identifier for the interface that consists
     * of the calling requirements for "function".
     */
    vrpsx_interface_identifier_t interface_identifier;

    /* A human-readable name for the interface that might help users
     * to understand the purpose or context of the interface identifier.
     *
     * WARNING: The display_name string is NOT an identifier.  The
     * same display name may be associated with several interfaces,
     * and different clients may choose different display names for
     * the same interface.  When reporting a display name, resolve
     * this ambiguity by mentioning the interface identifier as well.
     */
    char const *display_name;
} vrpsx_identified_function_t;


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_IDENTIFIED_FUNCTION_H */
